"""Утилиты для работы с сетевой конфигурацией."""
import json
import subprocess
from typing import Any, Callable, Optional, cast


NETPLAN_DEFAULT_ETH1_CONFIG = """network:
  version: 2
  ethernets:
    eth0:
      addresses:
      - 2.3.4.5/8
"""

NETPLAN_TEMPLATE_ETH = """network:
  version: 2
  ethernets:
    {iface}:
"""

NETPLAN_TEMPLATE_IP = """      addresses:
      - {ip_with_netmask}
"""

NETPLAN_TEMPLATE_GATEWAY = """      routes:
      - to: 0.0.0.0/0
        via: {gateway}
"""

NETPLAN_TEMPLATE_DNS = """      nameservers:
        addresses:
        - {ip_with_netmask}
"""


def get_runtime_iface_ip_address(ifname: str) -> str:
    """Получить текущий ip адрес интерфейса.

    Args:
        ifname (str): Имя интерфейса.

    Raises:
        ValueError: Возвращается если не найдено не одного адреса у интерфейса.

    Returns:
        str: IP адрес интерфейса с длиной маски. Пример 192.168.0.205/24.
    """
    command: list[str] = [
        'ip',
        '--json',
        'address',
        'show',
        ifname,
    ]
    res = subprocess.run(command, capture_output=True)
    iface_settings = json.loads(res.stdout)[0]
    for family in iface_settings['addr_info']:
        if family['family'] == 'inet':
            return f"{family['local']}/{family['prefixlen']}"
    raise ValueError(f'No ip address configured on the {ifname} interface.')


def get_runtime_iface_gateway(ifname: str) -> Optional[str]:
    """Получить текущий шлюз для интерфейса.

    Args:
        ifname (str): Имя интерфейса.

    Raises:
        ValueError: Возвращается если шлюз для интерфейса не сконфигурирован.

    Returns:
        str: IP адрес шлюза.
    """
    command: list[str] = [
        'ip',
        '--json',
        'route',
        'show',
        'dev',
        ifname,
        'match',
        '0/0',
    ]
    res = subprocess.run(command, capture_output=True)
    routes = json.loads(res.stdout)
    for route in routes:
        return cast(str, route.get('gateway'))
    return None


def get_runtime_iface_link_settings(ifname: str) -> dict[str, Any]:
    """Получить текущий ip адрес интерфейса.

    Args:
        ifname (str): Имя интерфейса.

    Raises:
        ValueError: Возвращается если не найдено не одного адреса у интерфейса.

    Returns:
        str: IP адрес интерфейса с длиной маски. Пример 192.168.0.205/24.
    """
    command: list[str] = [
        'ip',
        '--json',
        'link',
        'show',
        'dev',
        ifname,
    ]
    res = subprocess.run(command, capture_output=True)
    try:
        res.check_returncode()
    except subprocess.CalledProcessError:
        return {}
    iface_settings: dict[str, Any] = json.loads(res.stdout)[0]
    return iface_settings


def get_runtime_iface_mac_address(ifname: str) -> Optional[str]:
    link_settings = get_runtime_iface_link_settings(ifname)
    return link_settings.get('address')


def generate_artnet_standard_ip(mac: str, oem: int) -> str:
    mac_int_repr = [int(s, 16) for s in mac.split(':')]
    oem_bytes_repr = oem.to_bytes(2, byteorder='big')
    return f'2.{(mac_int_repr[3] + oem_bytes_repr[0] + oem_bytes_repr[1]) % 255}.{mac_int_repr[4]}.{mac_int_repr[5]}/8'


def make_eth_netplan_config_with_ip(ifname: str, ip_with_netmask: str) -> str:
    config = NETPLAN_TEMPLATE_ETH.format(iface=ifname)
    config += NETPLAN_TEMPLATE_IP.format(ip_with_netmask=ip_with_netmask)
    return config


def make_netplan_config_with_default_dns_from_runtime_settings(
    ifname: str = 'eth0',
    dns: str = '8.8.8.8',
    get_iface_ip: Callable[[str], str] = get_runtime_iface_ip_address,
    get_iface_gateway: Callable[[str], Optional[str]] = get_runtime_iface_gateway,
) -> str:
    """Подготавливает netplan конфигурацию интерфейса.

    Args:
        ifname (str, optional): Имя интерфейса. Defaults to 'eth0'.

    Returns:
        str: Netplan конфигурация интерфейса.
    """
    iface_ip_address = get_iface_ip(ifname)
    config = NETPLAN_TEMPLATE_ETH.format(iface=ifname)
    config += NETPLAN_TEMPLATE_IP.format(ip_with_netmask=iface_ip_address)
    if gateway := get_iface_gateway(ifname):
        config += NETPLAN_TEMPLATE_GATEWAY.format(gateway=gateway)
    config += NETPLAN_TEMPLATE_DNS.format(dns=dns)
    return config


def write_netplan_config(config: str, path: str = '10-ethernets.yaml') -> None:
    """Сохраняет netplan конфигурацию интерфейса на диск.

    Args:
        config (str): Netplan конфигурация интерфейса.
        path (str, optional): Имя файла конфигурации. Defaults to '10-ethernets.yaml'.
    """
    with open(path, 'w') as file:
        file.write(config)
