"""Entry point."""
import glob
import logging
import logging.config
import argparse
import os
import subprocess
import json
import pathlib
import shutil
import tarfile
import time
import traceback
import sqlite3
from typing import Any, Literal, Optional, Protocol, Type, TypedDict, Union

from update.log_config import LOGGING_CONFIG
from update.network_parser import (
    NETPLAN_DEFAULT_ETH1_CONFIG,
    generate_artnet_standard_ip,
    get_runtime_iface_mac_address,
    make_eth_netplan_config_with_ip,
    make_netplan_config_with_default_dns_from_runtime_settings,
    write_netplan_config,
)

logging.config.dictConfig(LOGGING_CONFIG)
logger = logging.getLogger(__name__)


RUNNING_PATH = pathlib.Path(os.path.realpath(__file__)).parent


def get_absolute_path(path: Union[str, pathlib.Path], parents_path: pathlib.Path) -> pathlib.Path:
    if pathlib.Path(path).is_absolute():
        return pathlib.Path(path)
    return parents_path / pathlib.Path(path)


module_updating_statuses = Literal[
    'prepared',
    'installing',
    'installed',
    'error',
    'restored',
]


PlayerVersion = Literal['1a', '1b', '2a']


class UpdatingCommandDict(TypedDict):
    """Updating command dict.

    command: str    - updating command.
    settings: Any   - settings for update.
    """

    command: str
    settings: Any


class UpdateModuleDict(TypedDict):
    """Updating module settings dict.

    module          - module name.
    version         - module version.
    update_stages   - list of updating stages.
    restore_stages - list of restoring stages.
    """

    module: str
    version: str
    player_versions: list[PlayerVersion]
    update_stages: list[UpdatingCommandDict]
    restore_stages: list[UpdatingCommandDict]


class MetadataDict(TypedDict):
    """Dict with settings for update.

    version             - whole version.
    update_modules      - list of updating modules.
    """

    version: str
    minimal_prev_version: str
    update_modules: list[UpdateModuleDict]
    empty_folders: list[str]


class ModuleStatusDict(TypedDict):
    """Dict with status of module updating process.

    module - module name.
    version - module version.
    status - updating status.
    """

    module: str
    version: str
    status: module_updating_statuses


class ModuleUpdatingStatusDict(TypedDict):
    """Dict with status of module updating process.

    module - module name.
    version - module version.
    status - updating status.
    """

    module: str
    status: module_updating_statuses


class CopyFilesSettingsDict(TypedDict):
    """НАстройки копирования файлов.

    src: str - source path.
    dst: str - destination path.
    excludes (list[str]): Исключаемые файлы.
    """

    src: Union[str, pathlib.Path]
    dst: Union[str, pathlib.Path]
    excludes: list[str]


class DeleteFileSettingsDict(TypedDict):
    """Replace file settings dict."""

    target: Union[str, pathlib.Path]
    sudo: bool
    check: bool


class EraseDirRequiredSettingsDict(TypedDict):
    """Обязательные поля для EraseDirRequiredSettingsDict."""

    target_dir: Union[str, pathlib.Path]
    sudo: bool


class EraseDirSettingsDict(EraseDirRequiredSettingsDict, total=False):
    """Replace file settings dict.

    target_dir (str): Путь до директории которую необходимо зачистить.
    sudo (bool): Нужно ли запускать команду от sudo.
    excludes (list[str]): Исключаемые файлы или каталоги. (Опционально)
    """

    excludes: list[str]


class ReplaceFileSettingsDict(TypedDict):
    """Replace file settings dict.

    src: str - source path.
    dst: str - destination path.
    skip_if_exist: bool - skip if file exist
    sudo: bool - run as sudo
    """

    src: str
    dst: str
    skip_if_exist: bool
    sudo: bool


class RestoreFileSettingsDict(TypedDict):
    """Replace file settings dict.

    src: str - source path.
    dst: str - destination path.
    sudo: Optional[bool] - run as sudo
    """

    src: str
    dst: str
    sudo: Optional[bool]


class ReplaceFileDependsOnPlayerVersionSettingsDict(TypedDict):
    """Replace file settings dict.

    src: dict[PlayerVersion, str]
    dst: dict[PlayerVersion, str]
    skip_if_exist: bool
    """

    src: dict[PlayerVersion, str]
    dst: dict[PlayerVersion, str]
    skip_if_exist: bool


class ReplaceFilesSettingsDict(TypedDict):
    """Replace file settings dict.

    src: str - source path.
    dst: str - destination path.
    excludes (list[str]): Исключаемые файлы.
    """

    src: str
    dst: str
    excludes: list[str]
    sudo: bool


class MoveFileSettingsDict(TypedDict):
    src: str
    dst: str
    sudo: bool
    check: bool
    skip_if_exist: bool


class InstallDependenciesSettingsDict(TypedDict):
    """Dict with settings for install python dependencies to venv.

    dependency_file: str    - path to requirements.txt.
    venv_dir: str           - path to venv.
    """

    dependency_file: str
    venv_dir: str


class InstallLocalDependenciesSettingsDict(TypedDict):
    """Dict with settings for install python dependencies to venv."""

    dependency_file: str
    path_to_dependency_archive: str
    venv_dir: str


class UninstallDependenciesSettingsDict(TypedDict):
    """Dict with settings for install python dependencies to venv.

    packages (list[str]): path to requirements.txt.
    venv_dir (str): path to venv.
    """

    packages: list[str]
    venv_dir: str


class CreatePythonVenvSettingsDict(TypedDict):
    """Dict with settings for creating python venv.

    venv_dir (str): path to venv.
    """

    venv_dir: str


class DbMigrateAlembicSettings(TypedDict):
    """Dict with settings for migration db with alembic.

    alembic_migration_dir: str  - path to alembic migrations.
    venv_dir: str               - path to venv.
    """

    migration_dir: str
    venv_dir: str


class IndependentDbMigrateAlembicSettings(TypedDict):
    """Dict with settings for migration db with alembic without python dependencies.

    alembic_migration_dir: str  - path to alembic migrations.
    """

    migration_dir: str


class DbMigrateDjangoSettings(TypedDict):
    """Dict with settings for migration db with django.

    django_app_path: str        - path to django app.
    venv_dir: str               - path to venv.
    license: str                - path to license file.
    migrate_script: str         - path to django migrate script.
    """

    django_app_path: str
    venv_dir: str
    license: str  # noqa A003
    migrate_script: str


class SqliteDbSettings(TypedDict):
    """Dict with settings for connect to sqlite db.

    db_path: str    - path to db.
    """

    db_path: str


class SystemdSettings(TypedDict):
    """Dict with settings for manipulating systemd service.

    service_name: str   - service name.
    check: bool         - check or not manipulating result.
    sudo: bool          - run as sudo
    """
    service_name: str
    check: bool
    sudo: bool


class SystemdReloadSettings(TypedDict):
    """Dict with settings for manipulating systemd service.

    check: bool         - check or not manipulating result.
    sudo: bool          - run as sudo
    """
    check: bool
    sudo: bool


class CheckableSettingsDict(TypedDict):
    """Словарь с настройкой check.

    check: bool         - check or not manipulating result.
    """

    check: bool


class DelaySettings(TypedDict):
    """Dict with settings for manipulating systemd service.

    delay: int   - delay in seconds.
    """
    delay: int


class CreateLinkSettings(TypedDict):
    """Dict with settings for link creation.

    src: str                - src file
    dst: str                - dst link
    is_symbol: bool         - is symbol link
    sudo: bool              - run as sudo
    """
    src: str
    dst: str
    is_symbol: bool
    sudo: bool


class InstallAptSettingsDict(TypedDict):
    """Replace file settings dict.

    apt_package_names (list[str]): Имена устанавливаемых apt пакетов.
    """

    apt_package_names: list[str]


class ShellCommandSettingsDict(TypedDict):

    command: str


class NetplanIfaceSettingsDict(TypedDict):
    """Словарь с настройками для создания конфигурации netplan.

    ifname (str): Имя сетевого интерфейса.
    netplan_config_filename (str): Имя файла конфигурации netplan. Пример 10-ethernets.yaml.
    """

    ifname: str
    netplan_config_filename: str


class UpdateStageHandleInterface(Protocol):
    """Interface for update stage handle command.

    Args:
        settings (Any): any settings for handling stage.
    """

    def __init__(self, settings: Any) -> None:
        ...

    def execute(self) -> None:
        """Run command."""
        ...


class RestoreStageHandleInterface(Protocol):
    """Interface for restore stage handle command.

    Args:
        settings (Any): any settings for handling stage.
        backup_dir (str): path to backup.
    """

    def __init__(self, settings: Any, backup_dir: str) -> None:
        ...

    def execute(self) -> None:
        """Run command."""
        ...


def custom_copy_fnc(src: str, dst: str, *args: Any, **kwargs: Any) -> str:
    """Копирует файл и в случае возникновения ошибки повторяет копирование с правами sudo.

    Функция предназначена для работы с shutil.copytree.
    """
    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    try:
        shutil.copy2(src, dst, *args, **kwargs)
    except BaseException as err:  # noqa B036
        if isinstance(err, OSError) and err.errno == 26 and err.filename.endswith('.venv/bin/python'):
            pass
        else:
            sudo_cp_command: list[str] = [
                'sudo',
                'cp',
                '-p',
                src,
                dst,
            ]
            subprocess.run(sudo_cp_command, check=True)
    return dst


class CopyFilesCommand():
    """Копирует файлы в определенный каталог.

    Args:
        settings (CopyFilesSettingsDict): Настройки для копирования файлов.
    """

    def __init__(self, settings: CopyFilesSettingsDict, *args: Any, **kwargs: Any) -> None:
        self.src = get_absolute_path(settings['src'], parents_path=RUNNING_PATH)
        self.dst = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH)
        self._excludes = settings.get('excludes', [])

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to copy files from %s to %s.', self.src, self.dst)
        logger.debug('Excludes: %s', self._excludes)
        ignore_fn = shutil.ignore_patterns(*self._excludes) if self._excludes else None
        shutil.copytree(self.src, self.dst, dirs_exist_ok=True, ignore=ignore_fn)
        logger.debug('Files has been copied from %s to %s.', self.src, self.dst)


class CopyFileIfNotExistCommand():
    """Копирует файл если он не существует.

    Args:
        settings (CopyFilesSettingsDict): Настройки для копирования файлов.
    """

    def __init__(self, settings: CopyFilesSettingsDict, *args: Any, **kwargs: Any) -> None:
        self.src = get_absolute_path(settings['src'], parents_path=RUNNING_PATH)
        self.dst = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH)

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to copy file from %s to %s.', self.src, self.dst)
        if not pathlib.Path(self.dst).exists():
            if not pathlib.Path(self.dst).parent.exists():
                pathlib.Path(self.dst).parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(self.src, self.dst)
            logger.debug('File has been copied from %s to %s.', self.src, self.dst)
        else:
            logger.debug('File already exist.')


class EraseDirCommand:
    """Очищает указанный каталог.

    Args:
        settings (EraseDirSettingsDict): Настройки для очистки каталога.
    """

    def __init__(self, settings: EraseDirSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._dir = settings['target_dir']
        self._sudo = settings['sudo']
        self._excludes = settings.get('excludes', [])

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to erase dir %s', self._dir)

        if not pathlib.Path(self._dir).exists():
            logger.debug('No such directory %s', self._dir)
            return

        for file in pathlib.Path(self._dir).iterdir():
            if file.name not in self._excludes:
                command: list[str] = [
                    'rm',
                    '-rf',
                    file.absolute().as_posix(),
                ]
                if self._sudo:
                    command.insert(0, 'sudo')
                res = subprocess.run(command)
                check_cmd_process(res)

        logger.debug('Dir %s erased', self._dir)


class DeleteFileCommand:
    """Удаляет указанный файл.

    Args:
        settings (DeleteFileSettingsDict): Настройки.
    """

    def __init__(self, settings: DeleteFileSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._target = pathlib.Path(settings['target'])
        self._check = settings['check']
        self._sudo = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to delete file %s', self._target)

        command: list[str] = [
            'rm',
            '-rf',
            self._target.absolute().as_posix(),
        ]

        if self._sudo:
            command.insert(0, 'sudo')

        res = subprocess.run(command)
        if self._check:
            check_cmd_process(res)

        logger.debug('File %s deleted', self._target)


class ReplaceFilesCommand:
    """Очищает каталог и копирует туда файлы из обновления.

    Args:
        settings (ReplaceFilesSettingsDict): Настройки для копирования файлов.
    """

    def __init__(self, settings: ReplaceFilesSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._src = get_absolute_path(settings['src'], parents_path=RUNNING_PATH)
        self._dst = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH)
        self._excluded = settings.get('excludes', [])
        self._sudo = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to replace files from %s to %s with excludes %s', self._src, self._dst, self._excluded)
        EraseDirCommand(settings={'target_dir': self._dst, 'sudo': self._sudo, 'excludes': self._excluded}).execute()
        CopyFilesCommand(
            settings={'src': self._src.absolute().as_posix(), 'dst': self._dst, 'excludes': self._excluded},
        ).execute()
        logger.debug('Files was replaced.')


class ReplaceFileCommand():
    """Copy ona file form src to dst.

    Args:
        settings (ReplaceFilesSettingsDict): _description_
    """

    def __init__(self, settings: ReplaceFilesSettingsDict, *args: Any, **kwargs: Any) -> None:
        self.src = get_absolute_path(settings['src'], parents_path=RUNNING_PATH)
        self.dst = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH)

    def execute(self) -> None:
        """Run command."""
        # copy files
        logger.debug('Preparing to copy file from %s to %s.', self.src, self.dst)
        shutil.copy2(self.src, self.dst)
        logger.debug('File has been copied from %s to %s.', self.src, self.dst)


class ReplaceFileDependsOnPlayerVersionCommand():
    """Copy ona file form src to dst.

    Args:
        settings (ReplaceFileDependsOnPlayerVersionSettingsDict): _description_
    """

    def __init__(self, settings: ReplaceFileDependsOnPlayerVersionSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._src = settings['src']
        self._dst = settings['dst']
        self._skip_if_exist = settings.get('skip_if_exist', False)

    def execute(self) -> None:
        """Run command."""
        player_version = get_player_version()
        if player_version not in self._src or player_version not in self._dst:
            raise Exception('Player version not found in settings.')

        src = RUNNING_PATH / self._src[player_version]
        dst = self._dst[player_version]

        logger.debug('Preparing to copy file from %s to %s.', src, dst)

        if self._skip_if_exist and pathlib.Path(dst).exists():
            logger.debug('File %s exists. Skip.', dst)
            return

        shutil.copy2(src, dst)

        logger.debug('File has been copied from %s to %s.', src, dst)


class ReplaceSystemFileCommand():
    """Copy ona file form src to dst with sudo.

    Копирует один файл с привилегиями sudo.
    Владелец файла после копирования становиться root.

    Args:
        settings (ReplaceFileSettingsDict): _description_
    """

    def __init__(self, settings: ReplaceFileSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._src = get_absolute_path(settings['src'], parents_path=RUNNING_PATH)
        self._dst = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH)
        self._skip_if_exist = settings['skip_if_exist']

    def execute(self) -> None:
        """Run command."""
        # copy files
        logger.debug('Preparing to copy system file from %s to %s.', self._src, self._dst)

        if self._skip_if_exist and pathlib.Path(self._dst).exists():
            logger.debug('File %s exists. Skip.', self._dst)
            return

        command: list[str] = [
            'sudo',
            'cp',
            self._src.absolute().as_posix(),
            self._dst.absolute().as_posix(),
        ]
        res = subprocess.run(command)
        check_cmd_process(res)
        logger.debug('System file has been copied from %s to %s.', self._src, self._dst)


class CopyFileCommand():
    """Copy one file form src to dst .

    Копирует один файл.
    Если используется sudo то владелец файла после копирования становиться root.

    Args:
        settings (ReplaceFileSettingsDict): _description_
    """

    def __init__(self, settings: ReplaceFileSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._src = get_absolute_path(settings['src'], parents_path=RUNNING_PATH)
        self._dst = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH)
        self._skip_if_exist = settings['skip_if_exist']
        self._sudo = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        # copy files
        logger.debug('Preparing to copy file from %s to %s.', self._src, self._dst)

        if self._skip_if_exist and pathlib.Path(self._dst).exists():
            logger.debug('File %s exists. Skip.', self._dst)
            return

        command: list[str] = [
            'cp',
            self._src.absolute().as_posix(),
            self._dst.absolute().as_posix(),
        ]
        if self._sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)
        check_cmd_process(res)
        logger.debug('File has been copied from %s to %s.', self._src, self._dst)


class MoveFileCommand:

    def __init__(self, settings: MoveFileSettingsDict, *args: Any, **kwargs: Any) -> None:
        src = pathlib.Path(settings['src'])
        self._src = get_absolute_path(src, parents_path=RUNNING_PATH)
        self._dst = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH)
        self._sudo = settings['sudo']
        self._check = settings['check']
        self._skip_if_exist = settings['skip_if_exist']

    def execute(self) -> None:
        """Run command."""
        # copy files
        logger.debug('Preparing to move file from %s to %s.', self._src, self._dst)

        if self._skip_if_exist and pathlib.Path(self._dst).exists():
            logger.debug('File %s exists. Skip.', self._dst)
            return

        command: list[str] = [
            'mv',
            self._src.absolute().as_posix(),
            self._dst.absolute().as_posix(),
        ]
        if self._sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)
        if self._check:
            check_cmd_process(res)
        logger.debug('File has been moved from %s to %s.', self._src, self._dst)


class RestoreFilesFromBackupCommand():
    """Replace all files in destination directory from backup.

    Args:
        settings (ReplaceFilesSettingsDict):    settings for replacing files.
        backup_dir (str):                       path to backup.
    """

    def __init__(self, settings: ReplaceFilesSettingsDict, backup_dir: str) -> None:
        self.src = get_absolute_path(settings['src'], parents_path=pathlib.Path(backup_dir))
        self.dst = settings['dst']

    def execute(self) -> None:
        """Replace all files in destination directory from backup."""
        logger.debug('Preparing to copy files from %s to %s.', self.src, self.dst)
        # erase dst dir
        shutil.rmtree(self.dst, ignore_errors=True)
        logger.debug('Files from dst directory %s has been erased.', self.dst)
        # copy files
        try:
            self._replace_files()
        except FileNotFoundError:
            logger.debug('Directory %s not found in the backup.', self.src)
            return
        logger.debug('Files has been copied from %s to %s.', self.src, self.dst)

    def _replace_files(self) -> None:
        shutil.copytree(self.src, self.dst, dirs_exist_ok=True, copy_function=custom_copy_fnc)


class RestoreOneFileFromBackupCommand():
    """Copy one files from backup ro destination.

    Args:
        settings (ReplaceFilesSettingsDict):    settings for replacing file.
        backup_dir (str):                       path to backup.
    """

    def __init__(self, settings: RestoreFileSettingsDict, backup_dir: str) -> None:
        self._src = get_absolute_path(settings['src'], parents_path=pathlib.Path(backup_dir))
        self._dst = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH)
        self._sudo = bool(settings.get('sudo', False))

    def execute(self) -> None:
        """Run command."""
        # copy file
        logger.debug('Preparing to restore file from %s to %s.', self._src, self._dst)
        """Run command."""

        if self._src.exists():
            command: list[str] = [
                'cp',
                self._src.absolute().as_posix(),
                self._dst.absolute().as_posix(),
            ]
            if self._sudo:
                command.insert(0, 'sudo')
            res = subprocess.run(command)
            check_cmd_process(res)
            logger.debug('File has been copied from %s to %s.', self._src, self._dst)
        else:
            logger.debug('File %s is not found in the backup.', self._src)
            if self._dst.exists():
                command = [
                    'rm',
                    self._dst.absolute().as_posix(),
                ]
                if self._sudo:
                    command.insert(0, 'sudo')
                res = subprocess.run(command)
                check_cmd_process(res)
                logger.debug('File %s has been removed.', self._dst)


class CreatePythonVenv:
    """Instal python dependencies.

    Args:
        settings (CreatePythonVenvSettingsDict): _description_
    """

    def __init__(self, settings: CreatePythonVenvSettingsDict, *args: Any, **kwargs: Any) -> None:
        self.venv_dir: str = settings['venv_dir']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to create python venv %s.', self.venv_dir)
        new_env = os.environ.copy()
        new_env["PYTHONPATH"] = ""
        create_venv_command: list[str] = [
            '/usr/bin/python',
            '-m',
            'venv',
            self.venv_dir,
            '--copies',
            '--clear',
            '--upgrade-deps',
        ]
        res = subprocess.run(create_venv_command, env=new_env)
        check_cmd_process(res)
        logger.debug('Venv has been created. %s', self.venv_dir)


class InstalPythonDependenciesCommand():
    """Instal python dependencies.

    Args:
        settings (InstallDependenciesSettingsDict): Настройки для установки python зависимостей.
    """

    def __init__(
        self,
        settings: InstallDependenciesSettingsDict,
        *args: Any,
        create_venv_command: Type[UpdateStageHandleInterface] = CreatePythonVenv,
        **kwargs: Any,
    ) -> None:
        self._requirements_file: str = str(RUNNING_PATH/settings['dependency_file'])
        self._venv_dir: str = settings['venv_dir']
        self._create_venv_command = create_venv_command({'venv_dir': self._venv_dir})

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to install python dependencies.')
        new_env = os.environ.copy()
        new_env["PYTHONPATH"] = ""
        # create venv
        if not pathlib.Path(self._venv_dir).exists():
            self._create_venv_command.execute()
        # install dependencies
        pip_path: str = self._venv_dir + '/bin/pip'
        dependencies_install_command: list[str] = [
            pip_path,
            'install',
            '-r',
            self._requirements_file,
        ]
        res = subprocess.run(dependencies_install_command, env=new_env)
        check_cmd_process(res)
        logger.debug('Dependencies has been installed to venv.')


class InstalLocalPythonDependenciesCommand():
    """Instal python dependencies.

    Args:
        settings (InstallDependenciesSettingsDict): Настройки для установки python зависимостей.
    """

    def __init__(
        self,
        settings: InstallLocalDependenciesSettingsDict,
        *args: Any,
        create_venv_command: Type[UpdateStageHandleInterface] = CreatePythonVenv,
        **kwargs: Any,
    ) -> None:
        self._requirements_file: str = str(RUNNING_PATH/settings['dependency_file'])
        self._archive_path: pathlib.Path = RUNNING_PATH/settings['path_to_dependency_archive']
        self._dependencies_path: pathlib.Path = pathlib.Path(self._archive_path).parent
        self._venv_dir: str = settings['venv_dir']
        self._create_venv_command = create_venv_command({'venv_dir': self._venv_dir})

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to install python dependencies.')

        # create venv
        self._create_venv_if_not_exists()

        # Untar archive
        self._untar_archive()

        # install dependencies
        self._install_dependencies()

        logger.debug('Dependencies has been installed to venv.')

    def _create_venv_if_not_exists(self) -> None:
        """Create venv if not exists."""
        if not pathlib.Path(self._venv_dir).exists():
            self._create_venv_command.execute()

    def _untar_archive(self) -> None:
        """Untar archive."""
        with tarfile.open(self._archive_path) as tar:
            tar.extractall(path=self._dependencies_path)

    def _install_dependencies(self) -> None:
        """Install dependencies."""
        new_env = os.environ.copy()
        new_env["PYTHONPATH"] = ""

        pip_path: str = self._venv_dir + '/bin/pip'
        dependencies_install_command: list[str] = [
            pip_path,
            'install',
            '--no-index',
            '--find-links',
            (self._dependencies_path / 'dependencies').absolute().as_posix(),
            '-r',
            self._requirements_file,
        ]
        res = subprocess.run(dependencies_install_command, env=new_env)
        check_cmd_process(res)


class UninstallPythonPackagesCommand():
    """Uninstall python packages.

    Args:
        settings (UninstallDependenciesSettingsDict): Настройки удаления пакетов python.
    """

    def __init__(self, settings: UninstallDependenciesSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._uninstall_packages = settings['packages']
        self._venv_dir = settings['venv_dir']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to uninstall python dependencies: %s.', ', '.join(self._uninstall_packages))
        new_env = os.environ.copy()
        new_env["PYTHONPATH"] = ""
        # install dependencies
        pip_path: str = self._venv_dir + '/bin/pip'
        command: list[str] = [
            pip_path,
            'uninstall',
            '-y',
            *self._uninstall_packages,
        ]
        res = subprocess.run(command, env=new_env)
        check_cmd_process(res)
        logger.debug('Dependencies %s uninstalled from venv %s.', ', '.join(self._uninstall_packages), self._venv_dir)


class LoadFixturesToDbDjangoCommand():
    """Migrate django db.

    Args:
        settings (ReplaceFilesSettingsDict): _description_
    """

    # TODO: Realise creation empty folder command.

    def __init__(self, settings: DbMigrateDjangoSettings, *args: Any, **kwargs: Any) -> None:
        self.django_app_path: str = settings['django_app_path']
        self.venv_dir: str = settings['venv_dir']
        self.license: str = settings['license']
        self.migrate_script = settings['migrate_script']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to load fixtures.')

        python_path: str = self.venv_dir + '/bin/python'
        modified_env = os.environ.copy()
        modified_env['PYARMOR_LICENSE'] = self.license

        shell_command: list[str] = [
            python_path,
            '-m',
            self.migrate_script,
            'loaddata',
            '*/fixtures/*.yaml',
        ]

        res = subprocess.run(
            shell_command,
            capture_output=True,
            cwd=self.django_app_path,
            env=modified_env,
        )
        check_cmd_process(res)
        logger.debug('Fixtures has been loaded.')


class MigrateDbDjangoCommand():
    """Migrate django db.

    Args:
        settings (ReplaceFilesSettingsDict): _description_
    """

    # TODO: Realeise creation empty folder command.

    def __init__(self, settings: DbMigrateDjangoSettings, *args: Any, **kwargs: Any) -> None:
        self.django_app_path: str = settings['django_app_path']
        self.venv_dir: str = settings['venv_dir']
        self.license: str = settings['license']
        self.migrate_script = settings['migrate_script']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to migrate db.')

        python_path: str = self.venv_dir + '/bin/python'
        modified_env = os.environ.copy()
        modified_env['PYARMOR_LICENSE'] = self.license

        shell_command: list[str] = [
            python_path,
            '-m',
            self.migrate_script,
            'migrate',
        ]

        res = subprocess.run(
            shell_command,
            capture_output=True,
            cwd=self.django_app_path,
            env=modified_env,
        )
        check_cmd_process(res)
        logger.debug('Db has been migrated.')


class IndependentMigrateDbAlembicCommand():
    """Migrate alembic db.

    Args:
        settings (ReplaceFilesSettingsDict): _description_
    """

    def __init__(self, settings: IndependentDbMigrateAlembicSettings, *args: Any, **kwargs: Any) -> None:
        self.alembic_migration_dir: pathlib.Path = RUNNING_PATH / settings['migration_dir']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to migrate db.')

        dependencies_install_command: list[str] = [
            'alembic',
            'upgrade',
            'head',
        ]

        res = subprocess.run(dependencies_install_command, cwd=self.alembic_migration_dir.absolute().as_posix())
        check_cmd_process(res)

        logger.debug('Db has been migrated.')


class MigrateDbAlembicCommand():
    """Migrate alembic db.

    Args:
        settings (ReplaceFilesSettingsDict): _description_
    """

    def __init__(self, settings: DbMigrateAlembicSettings, *args: Any, **kwargs: Any) -> None:
        self.alembic_migration_dir: str = settings['migration_dir']
        self.venv_dir: str = settings['venv_dir']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to migrate db.')

        python_path: str = self.venv_dir + '/bin/python'

        dependencies_install_command: list[str] = [
            python_path,
            '-m',
            'alembic',
            'upgrade',
            'head',
        ]

        res = subprocess.run(dependencies_install_command, cwd=self.alembic_migration_dir)
        check_cmd_process(res)

        logger.debug('Db has been migrated.')


class ReloadSystemdCommand():
    """Run command "systemctl daemon-reload".

    Args:
        settings (SystemdReloadSettings): _description_
    """

    def __init__(self, settings: SystemdReloadSettings, *args: Any, **kwargs: Any) -> None:
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to reload services %s')
        command: list[str] = [
            'systemctl',
            'daemon-reload',
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)
        if self.check:
            check_cmd_process(res)

        logger.debug('%s services reloaded')


class EnableSystemdServiceCommand():
    """Enable systemd service.

    Args:
        settings (SystemdSettings): _description_
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to enabled service %s', self.service_name)
        command: list[str] = [
            'systemctl',
            'enable',
            self.service_name,
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)
        if self.check:
            check_cmd_process(res)

        logger.debug('%s service has been enabled', self.service_name)


class DisableSystemdServiceCommand():
    """Disable systemd service.

    Args:
        settings (SystemdSettings): Параметры для отключения systemd сервиса.
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to disable service %s', self.service_name)
        command: list[str] = [
            'systemctl',
            'disable',
            self.service_name,
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)
        if self.check:
            check_cmd_process(res)

        logger.debug('%s service disabled', self.service_name)


class StopSystemdServiceCommand():
    """Stop systemd service.

    Args:
        settings (SystemdSettings): _description_
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to stop service %s', self.service_name)
        command: list[str] = [
            'systemctl',
            'stop',
            self.service_name,
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)

        if self.check:
            check_cmd_process(res)

        logger.debug('%s service has been stopped', self.service_name)


class RestartSystemdServiceCommand():
    """Restart systemd service.

    Args:
        settings (SystemdSettings): _description_
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to restart service %s', self.service_name)
        command: list[str] = [
            'systemctl',
            'restart',
            self.service_name,
        ]
        if self.sudo:
            command.insert(0, 'sudo')
        res = subprocess.run(command)

        if self.check:
            check_cmd_process(res)

        logger.debug('%s service has been restarted', self.service_name)


class StartSystemdServiceCommand():
    """Start systemd service.

    Args:
        settings (SystemdSettings): _description_
    """

    def __init__(self, settings: SystemdSettings, *args: Any, **kwargs: Any) -> None:
        self.service_name: str = settings['service_name']
        self.check: bool = settings['check']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to start service %s', self.service_name)

        command: list[str] = [
            'systemctl',
            'start',
            self.service_name,
        ]
        if self.sudo:
            command.insert(0, 'sudo')

        res = subprocess.run(command, check=False)

        if self.check:
            check_cmd_process(res)

        logger.debug('%s service has been started', self.service_name)


class CreateLinkCommand():
    """Creates file link.

    Args:
        settings (CreateLinkSettings): settings for link creation.
    """

    def __init__(self, settings: CreateLinkSettings, *args: Any, **kwargs: Any) -> None:
        self.src = get_absolute_path(settings['src'], parents_path=RUNNING_PATH).absolute().as_posix()
        self.dst: str = get_absolute_path(settings['dst'], parents_path=RUNNING_PATH).absolute().as_posix()
        self._is_symbol = settings['is_symbol']
        self.sudo: bool = settings['sudo']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to create link from file %s to %s', self.src, self.dst)

        if self._is_symbol:
            command: list[str] = ['ln', '-s', '-f', self.src, self.dst]
        else:
            command = ['ln', '-f', self.src, self.dst]
        if self.sudo:
            command.insert(0, 'sudo')
        subprocess.run(command, check=False)

        logger.debug('File link created.')


class RestartUpdaterServiceCommand():
    """Restart update service with delay."""

    def __init__(self, settings: DelaySettings, *args: Any, **kwargs: Any) -> None:
        self._delay: int = settings['delay']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Schedule to restart service lm_update.service after %s seconds', self._delay)

        delay_after_stop: int = 30

        command_stop: list[str] = [
            'sudo',
            'systemd-run',
            f'--on-active={self._delay}',
            '--timer-property=AccuracySec=100ms',
            'sudo',
            'systemctl',
            'stop',
            'lm_update.service',
        ]

        command_start: list[str] = [
            'sudo',
            'systemd-run',
            f'--on-active={self._delay + delay_after_stop}',
            '--timer-property=AccuracySec=100ms',
            'sudo',
            'systemctl',
            'start',
            'lm_update.service',
        ]

        subprocess.run(command_stop, check=True)
        subprocess.run(command_start, check=True)

        logger.debug('Schedule to restart service lm_update.service has been set.')


class PopulateUpdaterDbVersionTable():
    """Populate Updater Version table.

    Args:
        settings (SqliteDbSettings): Dict with settings for connect to sqlite db.
    """

    metadata: MetadataDict = None

    def __init__(self, settings: SqliteDbSettings, *args: Any, **kwargs: Any) -> None:
        self.path_to_db: str = settings['db_path']

    def execute(self) -> None:
        """Run command."""
        con = sqlite3.connect(self.path_to_db)
        cur = con.cursor()

        # Проверяем, существует ли запись
        cur.execute("SELECT * FROM current_version WHERE module = 'general'")
        is_exists = cur.fetchone()
        # если запись не существует то добавляем
        if not is_exists:
            insert_query: str = "INSERT INTO current_version (version, module) VALUES (?, ?)"
            cur.execute(insert_query, (self.metadata['version'], 'general'))
            con.commit()
        cur.close()
        con.close()


class DisableRedisPersistenceCommand():
    """Отключает redis persistence."""

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to disable redis persistence.')
        command: list[str] = [
            'sudo',
            'bash',
            './files/conf/disable_redis_persistence.sh',
        ]
        res = subprocess.run(command)
        check_cmd_process(res)
        logger.debug('Redis persistence disable.')


class InstallAptPackagesCommand():
    """Устанавливает apt пакет.

    Args:
        settings (InstallAptSettingsDict): Настройки для установки apt пакета.
    """

    def __init__(self, settings: InstallAptSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._apt_package_name = settings['apt_package_names']

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to install apt packages: %s.', ', '.join(self._apt_package_name))
        command: list[str] = [
            'sudo',
            'apt',
            'install',
            '-y',
            *self._apt_package_name,
        ]
        res = subprocess.run(command)
        check_cmd_process(res)
        logger.debug('Apt packages %s installed.', ', '.join(self._apt_package_name))


class RunShellCommand():

    def __init__(self, settings: ShellCommandSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._command = settings['command']

    def execute(self) -> None:
        """Run command."""
        logger.debug("Preparing to start command: '%s'", self._command)
        res = subprocess.run(self._command.split(' '), cwd=RUNNING_PATH)
        check_cmd_process(res)
        logger.debug("Command '%s' finished", self._command)


class UpdateAptRepositoriesCommand():
    """Обновляет apt репозитории."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    def execute(self) -> None:
        """Run command."""
        logger.debug('Preparing to update apt repositories.')
        command: list[str] = [
            'sudo',
            'apt',
            'update',
            '-y',
        ]
        res = subprocess.run(command)
        check_cmd_process(res)
        logger.debug('Apt repositories updated.')


class UpdateAptRepositoriesAndInstallPackages:
    """Обновляет apt репозиторий и устанавливает пакеты.

    Args:
        settings (InstallAptSettingsDict): Настройки для установки apt пакетов.
    """

    def __init__(self, settings: InstallAptSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._settings = settings

    def execute(self) -> None:
        """Run command."""
        UpdateAptRepositoriesCommand().execute()
        InstallAptPackagesCommand(self._settings).execute()


class MakeEhtNetplanIfaceConfigFromRuntimeWithDefaultDNS:
    """Создает файл с netplan конфигурацией интерфейса.

    Args:
        settings (NetplanIfaceSettingsDict): Настройки для создания netplan конфигурации.
    """

    def __init__(self, settings: NetplanIfaceSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._ifname = settings['ifname']
        self._netplan_config_filename = RUNNING_PATH / settings['netplan_config_filename']

    def execute(self) -> None:
        """Run command."""
        config = make_netplan_config_with_default_dns_from_runtime_settings(self._ifname)
        write_netplan_config(config, self._netplan_config_filename.absolute().as_posix())


class MakeEhtNetplanIfaceConfigWithArtNetIP:

    def __init__(self, settings: NetplanIfaceSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._ifname = settings['ifname']
        self._netplan_config_filename = RUNNING_PATH / settings['netplan_config_filename']

    def execute(self) -> None:
        """Run command."""
        mac = get_runtime_iface_mac_address(self._ifname)
        if mac:
            ip = generate_artnet_standard_ip(mac=mac, oem=0xffff)
            config = make_eth_netplan_config_with_ip(self._ifname, ip)
        else:
            config = NETPLAN_DEFAULT_ETH1_CONFIG
        write_netplan_config(config, self._netplan_config_filename.absolute().as_posix())


class ApplyNetplanConfig:
    """Применяет конфигурацию netplan.

    Args:
        settings (CheckableSettingsDict): Настройки для создания netplan конфигурации.
    """

    def __init__(self, settings: CheckableSettingsDict, *args: Any, **kwargs: Any) -> None:
        self._check = settings['check']

    def execute(self) -> None:
        """Run command."""
        """Run command."""
        logger.debug('Preparing to apply netplan config.')
        command: list[str] = [
            'sudo',
            'netplan',
            'apply',
        ]
        res = subprocess.run(command)
        if self._check:
            check_cmd_process(res)
        logger.debug('Netplan configuration was applied successfully.')


def create_state(metadata: MetadataDict) -> list[ModuleStatusDict]:
    """Create initial updating state.

    Creates list with information about updating modules.
    Information contain module name and updating status.

    Args:
        metadata (MetadataDict): Updating configuration.

    Returns:
        list[ModuleStatusDict]: list with information about updating modules.
    """
    state: list[ModuleStatusDict] = []
    for module in metadata['update_modules']:
        module_state = ModuleStatusDict(
            module=module['module'],
            version=module['version'],
            status='prepared',
        )
        state.append(module_state)
    return state


def create_module_state(module: UpdateModuleDict) -> ModuleStatusDict:
    module_state = ModuleStatusDict(
        module=module['module'],
        version=module['version'],
        status='installing',
    )
    return module_state


def update_state(state: list[ModuleStatusDict], new_module_state: ModuleUpdatingStatusDict) -> None:
    """Update information about certain updating module.

    Args:
        state (list[ModuleStatusDict]): Current updating state.
        new_module_state (ModuleUpdatingStatusDict): New module state.
    """
    for module_state_current in state:
        if module_state_current['module'] == new_module_state['module']:
            module_state_current['status'] = new_module_state['status']


def change_module_state(module: str, status: module_updating_statuses,  state: list[ModuleStatusDict]) -> None:
    """Update state for certain module.

    Args:
        module (str):                           name of updating module.
        status (module_updating_statuses):      updating status.
        state (list[ModuleStatusDict]):  updating state.
    """
    new_module_state: ModuleUpdatingStatusDict = {
        'module': module,
        'status': status,
    }
    update_state(state, new_module_state)


def save_state_to_file(state: list[ModuleStatusDict], filename: str) -> None:
    """Save updating state to file.

    Save updating state to file in JSON format.

    Args:
        state (list[ModuleStatusDict]): Current updating state.
        filename (str): path to file for saving state.
    """
    with open(filename, 'w') as file:
        json.dump(state, file, indent=4)
    logger.info('Updating state has been saved.')


update_stage_command_fabric: dict[str, Type[UpdateStageHandleInterface]] = {
    'replace_files': ReplaceFilesCommand,
    'replace_file': ReplaceFileCommand,
    'replace_file_depends_on_version': ReplaceFileDependsOnPlayerVersionCommand,
    'install_dependency': InstalPythonDependenciesCommand,
    'install_dependency_from_local': InstalLocalPythonDependenciesCommand,
    'migrate_db_django': MigrateDbDjangoCommand,
    'migrate_db_alembic': MigrateDbAlembicCommand,
    'independent_migrate_db_alembic': IndependentMigrateDbAlembicCommand,
    'start_service': StartSystemdServiceCommand,
    'stop_service': StopSystemdServiceCommand,
    'restart_service': RestartSystemdServiceCommand,
    'enable_service': EnableSystemdServiceCommand,
    'disable_service': DisableSystemdServiceCommand,
    'reload_services': ReloadSystemdCommand,
    'create_link': CreateLinkCommand,
    'restore_files': RestoreFilesFromBackupCommand,
    'restore_file': RestoreOneFileFromBackupCommand,
    'restart_updater': RestartUpdaterServiceCommand,
    'populate_updater_db': PopulateUpdaterDbVersionTable,
    'install_apt_packages': UpdateAptRepositoriesAndInstallPackages,
    'copy_files': CopyFilesCommand,
    'copy_file': CopyFileCommand,
    'move_file': MoveFileCommand,
    'delete_file': DeleteFileCommand,
    'copy_file_if_not_exist': CopyFileIfNotExistCommand,
    'uninstall_python_packages': UninstallPythonPackagesCommand,
    'replace_system_file': ReplaceSystemFileCommand,
    'erase_dir': EraseDirCommand,
    'make_runtime_iface_config': MakeEhtNetplanIfaceConfigFromRuntimeWithDefaultDNS,
    'generate_eth_netplan_config_with_artnet_ip': MakeEhtNetplanIfaceConfigWithArtNetIP,
    'apply_network_config': ApplyNetplanConfig,
    'load_fixtures_to_django_db': LoadFixturesToDbDjangoCommand,
    'disable_redis_persistance': DisableRedisPersistenceCommand,
    'run_shell_command': RunShellCommand,
}


restore_stage_command_fabric: dict[str, Type[RestoreStageHandleInterface]] = {
    'restore_files': RestoreFilesFromBackupCommand,
    'restore_file': RestoreOneFileFromBackupCommand,
    'reload_services': ReloadSystemdCommand,
    'start_service': StartSystemdServiceCommand,
    'stop_service': StopSystemdServiceCommand,
    'disable_service': DisableSystemdServiceCommand,
    'restart_updater': RestartUpdaterServiceCommand,
    'run_shell_command': RunShellCommand,
    'replace_system_file_from_backup': ReplaceSystemFileCommand,
    'copy_file': CopyFileCommand,
    'replace_file': ReplaceFileCommand,
    'delete_file': DeleteFileCommand,
}


def existing_path_type(value: str) -> str:
    """Check that path is exist.

    Implement type for argparse. Checks that value is existing path.

    Args:
        value (str): path to dir

    Raises:
        argparse.ArgumentTypeError: Raises if path is not exist.

    Returns:
        str: path to dir
    """
    if not pathlib.Path(value).exists():
        raise argparse.ArgumentTypeError('Path is not exist')
    return value


def update_module(
    updating_module: UpdateModuleDict,
    update_stage_command_fabric: dict[str, Type[UpdateStageHandleInterface]],
) -> None:
    """Update module.

    Args:
        updating_module (UpdateModuleDict):     _description_
        update_stage_command_fabric:            fabric with stage handle commands.
    """
    logger.info('Updating module %s', updating_module['module'])

    for updating_stage in updating_module['update_stages']:
        logger.debug('Start command: %s', updating_stage['command'])
        stage_command = update_stage_command_fabric[updating_stage['command']]
        stage_command(updating_stage['settings']).execute()
        logger.debug('Command %s has been finished successful.', updating_stage['command'])
        time.sleep(1)

    logger.info('%s module has been updated successfully.', updating_module['module'])


def restore_module(
    restoring_module: UpdateModuleDict,
    backup_dir: str,
    restore_stage_command_fabric: dict[str, Type[RestoreStageHandleInterface]],
) -> None:
    """Restore module.

    Args:
        restoring_module (UpdateModuleDict):     Updating module settings.
        backup_dir (str):                        path to backup.
        restore_stage_command_fabric:            fabric with stage handle commands.
    """
    logger.info('Restoring module %s', restoring_module['module'])

    for restoring_stage in restoring_module['restore_stages']:
        logger.debug('Start command: %s', restoring_stage['command'])
        stage_command = restore_stage_command_fabric[restoring_stage['command']]
        stage_command(restoring_stage['settings'], backup_dir).execute()
        logger.debug('Command %s has been finished successful', restoring_stage['command'])
        time.sleep(1)

    logger.info('%s module has been restored.', restoring_module['module'])


def check_cmd_process(cmd_process: subprocess.CompletedProcess[bytes]) -> None:
    """Check that cmd process return without problem.

    Check that return code is zero or logging stderr and raises error.

    Args:
        cmd_process (subprocess.CompletedProcess): Complete command process.

    Raises:
        subprocess.CalledProcessError: raise if return code is not zero.
    """
    if cmd_process.returncode != 0:
        if cmd_process.stderr:
            for row in cmd_process.stderr.decode().splitlines():
                logger.error(row)
        raise subprocess.CalledProcessError(
            cmd_process.returncode,
            cmd_process.args,
            output=cmd_process.stdout,
            stderr=cmd_process.stderr,
        )


def get_player_version() -> PlayerVersion:
    """Reboot device."""
    interfaces = glob.glob("/dev/ttyACM*")
    version: PlayerVersion
    if len(interfaces) == 4:
        version = '2a'
    elif len(interfaces) == 2:
        version = '1b'
    elif len(interfaces) == 0:
        version = '1a'
    else:
        raise ValueError("unsupported player version")
    return version


def main() -> None:
    """MAIN."""
    logger.info('Preparing to install update.')

    player_version = get_player_version()

    with open(RUNNING_PATH/'metadata.json') as file:
        metadata: MetadataDict = json.load(file)
    logger.debug('Metadata has been loaded.')

    # FIXME: Костыль. Подумать как исправить.
    PopulateUpdaterDbVersionTable.metadata = metadata

    parser = argparse.ArgumentParser(description=f"Update lightmaster player to version {metadata['version']}.")
    parser.add_argument('-b', '--backup', type=existing_path_type, required=False)
    args = parser.parse_args()
    backup_dir: str = args.backup
    if backup_dir:
        logger.debug('Backup dir has been received.')

    state: list[ModuleStatusDict] = []
    save_state_to_file(state, str(RUNNING_PATH/'STATE.json'))

    logger.info('Installing update.')
    try:
        for folder in metadata['empty_folders']:
            pathlib.Path(folder).mkdir(parents=True, exist_ok=True)
        for updating_module in metadata['update_modules']:
            if player_version not in updating_module['player_versions']:
                continue
            state.append(create_module_state(updating_module))
            try:
                update_module(updating_module, update_stage_command_fabric)
                change_module_state(updating_module['module'], 'installed', state)
            except BaseException as err:  # noqa B036
                logger.error(traceback.format_exc())
                logger.exception(
                    'Error has been occurred during updating module %s wit err: %s',
                    updating_module['module'], err,
                )
                change_module_state(updating_module['module'], 'error', state)

                if backup_dir:
                    logger.info('Start to restore from backup.')
                    modules_for_restoring = [module['module'] for module in state if module['status'] != 'prepared']
                    for updating_module in metadata['update_modules']:
                        if updating_module['module'] in modules_for_restoring:
                            restore_module(updating_module, backup_dir, restore_stage_command_fabric)
                            change_module_state(updating_module['module'], 'restored', state)
                break
    except BaseException as err:
        raise err
    finally:
        save_state_to_file(state, str(RUNNING_PATH/'STATE.json'))


if __name__ == '__main__':
    main()
